/*

 B I B 2 R I S . C

 Converts files from VCH BIBLIO 2.2 "Archive" format to "RIS".
 
 This works only with files from BIBLIO 2.2 (yes, the DOS version 
 from 1991!). Within BIBLIO 2.2, such an archive can be created under
 the "Organisation" menu. 

 The "export" function of the 16-bit WinBib 3.x-Versions is broken 
 (e.g. omits Patent numbers, bails out with long texts, etc.), and 
 as VCH did never manage to deliver the update to the 32-bit version 
 I decided to write my own export routine. 

 --------------------------------------------------------------------

 Copyright (C) 2002 Joerg Hau. All rights reserved.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License as
 published by the Free Software Foundation; either version 2 of
 the License, or (at your option) any later version. See the
 file LICENSE for details.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 --------------------------------------------------------------------

 Author:        Joerg Hau

 email:         joerg.hau(at)dplanet.ch
 www:           http://www.mysunrise.ch/users/joerg.hau/

 Creation:      2002-08-22 by JHa

 Modifications: 2008-08-29, JHa (functional)
                2002-09-03, JHa (rearrange author names)
                2002-09-06, JHa (RP field)

                (adapt VERSION below when changing!)

 This should compile with any C compiler,
 e.g. "gcc bib2ris.c -o bib2ris -O3 -Wall"

 To run: "./bib2ris my-biblio.a1 > my-biblio.txt" 

 To use the resulting file unter MS-Windows, you need to convert 
 the charset from MS-DOS (in my case, CP 437) to ANSI:  

     "recode ibm437/..ms-ansi my-biblio.txt"

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>              /* 'tolower()' */

/* --- some constants --- */

#define VERSION "V20020906"     /* String! */
#define MAXLEN 30000            /* max length of a field */
#define arraysize(x) (sizeof(x)/sizeof(x[0]))

/* BIBLIO knows only three categories: Journal, Patent, Book ...
   all that does not fit in the first two categories is
   essentially a book-like publication ;-)    */

#define JOURNAL 0
#define BOOK    1
#define PATENT  2

#define NUM_FIELDS 14           /* possible fields in a BIBLIO db */


/* --- Global variables --- */

const char *reftype[] = {"JOUR", "SER", "PAT"};         /* RIS codes for categories */
const char *field[][NUM_FIELDS] = {
        {"TI", "AU", "JO", "VL", "PY", "IS", "SP", "AB", "KW", "RP", "AV", "U1", "AD", "U5"},
        {"TI/T3", "AU/ED", "PB/CY", "VL/N1", "PY/ET", "SN", "SP", "AB", "KW", "RP", "AV", "U1", "AD", "U5"},
        {"TI", "AU/ED", "CY/IS", "M1", "PY", "CY", "PB", "AB", "KW", "RP", "AV", "U1", "AD", "U5"} };

char *buf;
FILE *fp;


/* --- function prototypes --- */

int arrange_authors (char *in);
int convert(const char *fnam);
int decode(const int type, char *buf, const int id);
int decode_multiple (char *txt, const char *tag);
int read_entry(void);
int recode_copy (char *txt, const char *tag);
int recode_date (const char *txt, const char *tag);
char *split_line (char *txt, const char *delim);


/************************************************************************
* Function:     arrange_authors                                         *
* Description:  re-writes namesfrom "First M. Last" to "Last, First M." *
*               removes also ':' char                                   *
* Arguments:    pointer to name                                         *
* Returns:      always 1                                                *
*************************************************************************/
int arrange_authors (char *in)
{
char *ptr = in;
int len;

/* if the string starts with '*', the name must no be changed
   (e.g. company or institution) ... only remove the '*' */

if (ptr[0] == '*')
	{
        memmove(ptr, ptr+1, strlen(ptr));
	printf("%s", ptr);
	return 1;
	}  

/* see if this is a name like "P. :van Here" */
ptr = strpbrk(in, ":=");

/* if this is not found, it must be a "normal" name,
   so look for the last space char */
if (!ptr)
   ptr = strrchr(in, ' ');

/* Still not found means it's a one-word-name, so reset 
   pointer to the *whole* line */
if (!ptr)
        ptr = in;

/* if the string has a space at its beginning, clean it up */
if (ptr[0] == ' ')
        memmove(ptr, ptr+1, strlen(ptr));

/* if the string has ':'  at its beginning, clean it up */
if (ptr[0] == ':')
        memmove(ptr, ptr+1, strlen(ptr));

/* now print it */              
printf("%s", ptr);  

if (strlen(ptr) == strlen(in))  /* was it at beginning? */
        return 1;               /* OK, finished */

/* else terminate in by inserting 0x0 at the end */
len = strlen(in) - strlen(ptr);
*(in+len) = 0x0;

printf (", %s", in);   /* print (beginning of) string */
return 1;
}


/************************************************************************
* Function:     convert                                                 *
* Description:  Reads & converts a given file                           *
* Arguments:    file name                                               *
* Returns:      number of entries that were converted                   *
*************************************************************************/
int convert(const char *fnam)
{
extern FILE *fp;
extern char *buf;
char ch;
int i, cnt = 0, type = JOURNAL;

if (!(fp = fopen(fnam, "rb")))          /* open file */
        {
        fprintf(stderr, "\n'convert': cannot open %s!", fnam);
        return 0;
        }

buf = (char *)calloc (MAXLEN+1, sizeof(char));  /* allocate buffer */
if (NULL == buf)
        {
        perror ("\'convert': cannot allocate buffer!");
        return 0;
        }

fprintf (stderr, "\nConverting %s ", fnam);

for (i = 0; i < 25; i++)        /* skip beginning of file */
        fgetc(fp);

while (!feof(fp))               /* read until eof */
        {
        /* read up to next entry. 
           Delimiters are "Z(eitschr), P(atent), B(uch) */
        do      ch = fgetc (fp);
                while ((ch != 'Z') && (ch != 'P') && (ch != 'B') && (!feof(fp)));

        /* just skip the rest of the description */
        for (i = 0; i < 10; i++)
                fgetc(fp);

        /* decode ID ... at this point the first char is already read */
        switch (ch)
                {
                case 'Z':
                        type = JOURNAL;
                        decode(type, buf, ++cnt);
                        break;
                case 'B':
                        type = BOOK;
                        decode(type, buf, ++cnt);
                        break;
                case 'P':
                        type = PATENT;
                        decode(type, buf, ++cnt);
                        break;
                default:
                        break;
                }
        /* entries are terminated with a series of 0x0 */
        for (i = 0; i < 20; i++)
                fgetc(fp);

	/* show progess: one dot every 100 entries */
	if (0 == (cnt%100))
		fprintf(stderr, ".");
        }

free (buf);
fclose(fp);
return cnt;
}



/************************************************************************
* Function:     decode                                                  *
* Description:  parses and "decodes" an entry                           *
* Arguments:    type of reference (JOURNAL; BOOK, PATENT), id number    *
* Returns:      1 if OK, 0 if problem.                                  *
*************************************************************************/
int decode(const int type, char *buf, const int id)
{
int i, is_split;
char *ch = NULL, tag[8] = "", *tag2 = NULL;

printf("\nTY  - %s", reftype[type]);    /* print type */
printf("\nID  - %d", id);               /* print ID number */
for (i = 0; i < NUM_FIELDS; i++)
        {
        read_entry();                   /* fill buffer */
        is_split = 0;                   /* flag if there is a "split" entry" */
        strcpy(tag, field[type][i]);    /* copy tag for easier work */

        /* if tag contains slash, split up */
        if (strchr(tag, '/'))
                {
                tag2 = split_line(tag, "/");

                /* "tag2" now points to second tag, "tag" to first */

                /* if ' / ' is found, split line */
                if ((ch = split_line(buf, " / ")) != NULL)
                        is_split = 1;   /* flag that split was found */
                }

        /* ... now parse the fields: */

        rerun:

        if (!strcmp(tag, "KW"))                 /* keywords */
                decode_multiple(buf, tag);
        else if (!strcmp(tag, "AU"))            /* author */
                decode_multiple(buf, tag);
        else if (!strcmp(tag, "ED"))            /* editor */
                decode_multiple(buf, tag);
        else if (!strcmp(tag, "U5"))            /* archive date from BIBLIO */
                recode_date(buf, tag);
        else if (!strcmp(tag, "RP"))            /* reprint goes here */
                recode_copy(buf, tag);
        else                                    /* all else */
                printf("\n%-3s - %s", tag, buf);

        if (is_split > 0)       /* re-run the whole thing if there was a split field */
                {
                strcpy (tag, tag2);     /* copy 2nd set of tags */
                strcpy (buf, ch);
                is_split = 0;           /* reset */
                goto rerun;             /* I hate this but ... */
                }
        }
printf("\nER  -\n");    /* end-of-reference tag */
return 1;
}


/************************************************************************
* Function:     decode_multiple                                         *
* Description:  parse fields with multiple keys (e.g. authors, keywords)*
*               ',' and ';' are separator chars                         *
* Arguments:    ptr to buffer to parse, ptr to label                    *
* Returns:      number of items                                         *
*************************************************************************/
int decode_multiple (char *txt, const char *tag)
{
const char delim[] = ";,";
char *ptr = txt, out[MAXLEN/10] = "";
int i = 0, j;

/* very special treatment for "full" names, marked by an asterisk */

if ((!strcmp(tag, "AU") || !strcmp(tag, "ED")) && (strstr(txt,"*")))
	{  
        printf("\n%-3s - ", tag);       /* print tag */
        memmove(txt, txt+1, strlen(txt));
	printf(txt);
	return 1;
	}

ptr = strtok(txt, delim);
while (ptr != NULL)
        {
        printf("\n%-3s - ", tag);       /* print first tag */
        strcpy (out, ptr);

        /* if the string has whitespace at its beginning, clean it up */
        if (out[0] == ' ')
                memmove(out, out+1, strlen(ptr));

        /* special treatment for author names */
        if (!strcmp(tag, "AU") || !strcmp(tag, "ED"))  
		arrange_authors(out);

        /* keywords should be in lowercase */
        else if (!strcmp(tag, "KW"))
                for (j = 0; j < strlen(out); j++)       
                        printf("%c", tolower(out[j]));

        /* everything else is just printed "as is" */
        else 
                printf(out);

        ptr = strtok (NULL, delim);             /* next token */
        i++;
        }
return i++;            /* number of items */
}


/************************************************************************
* Function:     read_entry                                              *
* Description:  Reads from file into buffer until 0x0 is encountered    *
* Arguments:    none                                                    *
* Returns:      length of string w/o NULL.                              *
* Remark:       uses global variables.                                  *
*************************************************************************/
int read_entry(void)
{
extern FILE *fp;
extern char *buf;
char ch;
int i = 0;

memset (buf, 0x0, MAXLEN);      /* clear buffer */

while ((ch = fgetc(fp)) != 0)
        buf[i++] = ch;

return i;       /* return length of string w/o terminating NULL */
}



/************************************************************************
* Function:     recode_date                                             *
* Description:  rewrites date from d.M.yyyy to yyyy/MM/dd               *
* Arguments:    ptr to buffer to parse, ptr to label                    *
* Returns:      1 if OK, 0 if not a valid date                          *
*************************************************************************/
int recode_date (const char *txt, const char *tag)
{
int year, month, day;

sscanf(txt, " %d. %d. %d", &day, &month, &year);

/* if there was just one item, sscanf may get weird results: */
if ((year < 1000) || (year > 2010))
        return 0;

/* this is almost ISO notation, except for the slashes ;-) */

/* The following line is RIS format, but is not imported anyway ... */
// printf("\n%-3s - %4d/%02d/%02d/", tag, year, month, day);

/* ... so let's stick with proper ISO 8601 formatting: */
printf("\n%-3s - %4d-%02d-%02d", tag, year, month, day);
return 1;
}



/************************************************************************
* Function:     recode_copy                                             *
* Description:  rewrites "COPY?" field                                  *
* Arguments:    ptr to buffer to parse, ptr to label                    *
* Returns:      1 if "in file", 0 if not                                *
* Note:         Reference Manager has a bug upon import: if the default *
*               reprint status is set to "In File", "RP" entries are    *
*               ignored. Thus, upon import, set "default reprint status"* 
*               to "Not In File".                                       *
*************************************************************************/
int recode_copy (char *txt, const char *tag)
{
if (strtok(txt, "yYjJ+"))       /* not found */
        {
        printf("\n%-3s - NOT IN FILE", tag);
        return 0;
        }

/* else = if found */

printf("\n%-3s - IN FILE", tag);
return 1;
}


/************************************************************************
* Function:     split_line                                              *
* Description:  split two fields in a line                              *
*               (BIBLIO uses e.g. "AU / ED" in the same field)          *
* Arguments:    ptr to buffer to parse, ptr to delimiter                *
* Returns:      ptr to substring FOLLOWING the delimiter                *
* IMPORTANT:    the buffer *txt is changed: cut off BEFORE the delimiter*
*************************************************************************/
char *split_line (char *txt, const char *delim)
{
char *ch, *ptr = txt;

ch = strstr(ptr, delim);
if (NULL == ch)
        return NULL;            /* return  if not found */

/* if substring was found, shorten original string
   by inserting 0x0 at th end */

*(txt+(strlen(txt)-strlen(ch))) = 0x0;

return (ch + strlen(delim));    /* ptr to substring AFTER delimiter */
}



/****************************************************************
* Function:     main -- Main program part.                      *
* Agruments:    Filename (incl. extension, via cmd line).       *
* Returns:      One of the following:                           *
*               0       OK                                      *
*               1       no file specified, or wrong option      *
*               4       internal error (calloc...)              *
****************************************************************/
int main(int argc, char *argv[])
{
static char *disclaimer =
"\nbib2ris - convert VCH BIBLIO 2.2 archive files to RIS format. %s.\n"
"Copyright (C) 2002 by Joerg Hau <joerg.hau@dplanet.ch>.\n\n"
"This program is free software; you can redistribute it and/or modify it under\n"
"the terms of the GNU General Public License as published by the Free Software\n"
"Foundation; either v2 of the License, or (at your option) any later version.\n\n"
"This program is distributed in the hope that it will be useful, but WITHOUT ANY\n"
"WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A\n"
"PARTICULAR PURPOSE. See the GNU General Public License for details.\n";

static char *msg=
"\nSyntax: bib2ris my-biblio.a1 [ | recode ibm437/..ms-ansi ] [ > my-biblio.txt ]\n\n";

static char *note=
"Note: If you import this file into Reference Manager, remember to set the\n" 
"'default reprint status' to 'Not In File'.\n";
                                      
int i = 0;

/* --- show the usual text --- */

fprintf(stderr, disclaimer, VERSION);

if (argv[1] == NULL)         /* no remaining parameter on cmd line? */
        {
        fprintf (stderr, msg);
        return 1;
        }

i = convert(argv[1]);

if (i == 0)
        {
        fprintf(stderr, "\n*** Program aborted.\n");
        return 2;
        }

fprintf (stderr, " success: %d entries.\n", i);
fprintf (stderr, "\n%s\n", note);
return 0;
}
